@@ -1,3 +1,4 @@
+from os import pipe
 from statefun_tasks.context import TaskContext
 from statefun_tasks.pipeline_impl.handlers import PipelineMessageHandler
 from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, TaskStatus, PipelineState, ChildPipeline, TaskInfo
@@ -15,7 +16,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
             and isinstance(message, TaskRequest) \
                 and not self.graph.is_empty()
 
-    def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline: '_Pipeline', task_state: Any):
+    async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline: '_Pipeline', task_state: Any):
         invoking_task = message  # type: TaskRequest
 
         # ensure we pick up the correct caller id when task producing this pipeline is a retry
@@ -53,7 +54,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
                 task.request = self._serialiser.serialise_args_and_kwargs(([]), {})
 
         # 3. split into tasks to call now and those to defer if max parallelism is exceeded
-        self.submitter.submit_tasks(context, tasks, max_parallelism=max_parallelism)
+        await self.submitter.submit_tasks(context, tasks, max_parallelism=max_parallelism)
 
         # 4. break
         return False, message
