@@ -1,3 +1,4 @@
+from statefun_tasks.storage import StorageBackend
 from statefun_tasks.serialisation import DefaultSerialiser
 from statefun_tasks.pipeline_builder import PipelineBuilder
 
@@ -42,6 +43,7 @@ class FlinkTasks(object):
         self._serialiser = serialiser if serialiser is not None else DefaultSerialiser()
         self._bindings = {}
         self._events = EventHandlers()
+        self._storage = None
 
         self.register_builtin('run_pipeline', _run_pipeline)
 
@@ -52,6 +54,84 @@ class FlinkTasks(object):
             ChildPipelineHandler()
         ]
 
+    @staticmethod
+    def extend(function, retry_policy: RetryPolicy = None, **params):
+        """
+        Adds the extensions to make a function unsable by a PipelineBuilder -> fn.send(), fn.to_task() and fn.defaults()
+
+        :param function: the function to wrap
+        :param retry_policy: retry policy to use should the task throw an exception
+        :param params: any additional parameters to the Flink Task (such as a retry policy)
+        """
+        def defaults():
+            return {
+                'retry_policy': None if retry_policy is None else retry_policy.to_proto(),
+                **params
+            }
+
+        def send(*args, **kwargs):
+            return PipelineBuilder().send(function, *args, **kwargs)
+
+        def to_task(args, kwargs, is_finally=False, parameters={}):
+            parameters = {**defaults(), **parameters}
+
+            if is_finally:
+                parameters['is_fruitful'] = False
+            
+            module_name = parameters.get('module_name', None)
+            task_type = _task_type_for(function, module_name)
+
+            args = _unpack_single_tuple_args(args)
+            return Task.from_fields(_gen_id(), task_type, args, kwargs, is_finally=is_finally, **parameters)
+
+        function.send = send
+        function.to_task = to_task
+        function.defaults = defaults
+
+        return function
+
+    def bind(
+        self, 
+        namespace: str = None, 
+        worker_name: str = None, 
+        retry_policy: RetryPolicy = None, 
+        with_state: bool = False, 
+        is_fruitful: bool = True, 
+        module_name: str = None,
+        with_context: bool = False,
+        display_name: str = None):
+        """
+        Decorator to bind a function as a Flink Task
+
+        :param namespace: namespace to use in place of the default
+        :param worker_name: worker name to use in place of the default
+        :param retry_policy: retry policy to use should the task throw an exception
+        :param with_state: whether to pass a state object as the first (second if with_context is also set) parameter.  The return value should be a tuple of state, result (default False)
+        :param is_fruitful: whether the function produces a fruitful result or simply returns None (default True)
+        :param module_name: if specified then the task type used in addressing will be module_name.function_name
+                            otherwise the Python module containing the function will be used
+        :param with_context: whether to pass a Flink context object as the first parameter (default false)
+        :param display_name: optional friendly name for this task
+        """
+        
+        def wrapper(function):
+
+            function = FlinkTasks.extend(function, 
+                    namespace = self._default_namespace if namespace is None else namespace,
+                    worker_name = self._default_worker_name if worker_name is None else worker_name,
+                    retry_policy = retry_policy,
+                    with_state = with_state,
+                    is_fruitful = is_fruitful,
+                    module_name = module_name,
+                    with_context = with_context,
+                    display_name = display_name
+            )
+
+            self.register(function, wrapper=None, **function.defaults())
+            return function
+
+        return wrapper
+
     @property
     def events(self) -> EventHandlers:
         """
@@ -59,6 +139,15 @@ class FlinkTasks(object):
         """
         return self._events
 
+    def set_storage_backend(self, storage: StorageBackend):
+        """
+        Sets the storage backend to use if required. The default is to use Flink state exclusively but temporary results from large parallelisms 
+        may cause memory issues in which case they can be saved outside in e.g. an S3 bucket or Redis
+
+        :param storage: results storage backend to use
+        """
+        self._storage = storage
+
     def value_specs(self):
         return self._value_specs
 
@@ -69,7 +158,7 @@ class FlinkTasks(object):
 
         :param type_name: task type to expose the built in as
         :param fun: a function, partial or lambda representing the built in
-        :param params: any additional parameters to the built in
+        :param params: any additional parameters to the Flink Task (such as a retry policy)
         """
         self._bindings[f'__builtins.{type_name}'] = FlinkTask(fun, self._serialiser, self._events, **params)
 
@@ -90,65 +179,6 @@ class FlinkTasks(object):
         fun.type_name = _task_type_for(fun, module_name)
         self._bindings[fun.type_name] = FlinkTask(wrapper or fun, self._serialiser, self._events, **params)
 
-    def bind(
-        self, 
-        namespace: str = None, 
-        worker_name: str = None, 
-        retry_policy: RetryPolicy = None, 
-        with_state: bool = False, 
-        is_fruitful: bool = True, 
-        module_name: str = None,
-        with_context: bool = False,
-        display_name: str = None):
-        """
-        Binds a function as a Flink Task
-
-        :param namespace: namespace to use in place of the default
-        :param worker_name: worker name to use in place of the default
-        :param retry_policy: retry policy to use should the task throw an exception
-        :param with_state: whether to pass a state object as the first (second if with_context is also set) parameter.  The return value should be a tuple of state, result (default False)
-        :param is_fruitful: whether the function produces a fruitful result or simply returns None (default True)
-        :param module_name: if specified then the task type used in addressing will be module_name.function_name
-                            otherwise the Python module containing the function will be used
-        :param with_context: whether to pass a Flink context object as the first parameter (default false)
-        :param display_name: optional friendly name for this task
-        """
-        def wrapper(function):
-            def defaults():
-                return {
-                    'namespace': self._default_namespace if namespace is None else namespace,
-                    'worker_name': self._default_worker_name if worker_name is None else worker_name,
-                    'retry_policy': None if retry_policy is None else retry_policy.to_proto(),
-                    'with_state': with_state,
-                    'is_fruitful': is_fruitful,
-                    'module_name': module_name,
-                    'with_context': with_context,
-                    'display_name': display_name
-                }
-
-            def send(*args, **kwargs):
-                return PipelineBuilder().send(function, *args, **kwargs)
-
-            def to_task(args, kwargs, is_finally=False, parameters={}):
-                parameters = {**defaults(), **parameters}
-
-                if is_finally:
-                    parameters['is_fruitful'] = False
-                
-                module_name = parameters.get('module_name', None)
-                task_type = _task_type_for(function, module_name)
-
-                args = _unpack_single_tuple_args(args)
-                return Task.from_fields(_gen_id(), task_type, args, kwargs, is_finally=is_finally, **parameters)
-
-            function.send = send
-            function.to_task = to_task
-
-            self.register(function, **defaults())
-            return function
-
-        return wrapper
-
     def get_task(self, task_type) -> FlinkTask:
         """
         Returns the Flink Task instance for a given task type name
@@ -161,20 +191,7 @@ class FlinkTasks(object):
         else:
             raise RuntimeError(f'{task_type} is not a registered FlinkTask')
 
-    def is_async_required(self, message: Union[TaskRequest, TaskResult, TaskException, TaskActionRequest, ChildPipeline]):
-        """
-        Checks if a task is declared as async or not
-
-        :param message: the task input protobuf message
-        :return: true if the task being called is declared as async otherwise false
-        """
-
-        try:
-            return isinstance(message, TaskRequest) and self.get_task(message.type).is_async
-        except:
-            return False
-
-    async def run_async(self, context: BatchContext, message: Union[TaskRequest, TaskResult, TaskException, TaskActionRequest, ChildPipeline]):
+    async def run(self, context: BatchContext, message: Union[TaskRequest, TaskResult, TaskException, TaskActionRequest, ChildPipeline]):
         """
         Runs an async Flink Task
 
@@ -189,10 +206,7 @@ class FlinkTasks(object):
                     try:
                         _log.info(f'Starting {task_context}')
 
-                        if self.is_async_required(message):
-                            await handler.handle_message_async(self, task_context, message)
-                        else:
-                            handler.handle_message(self, task_context, message)
+                        await handler.handle_message(self, task_context, message)
 
                         _log.info(f'Finished {task_context}')
 
@@ -205,36 +219,6 @@ class FlinkTasks(object):
             
             _log.error(f'Unsupported message type {message.typed_value.typename}')
 
-
-    def run(self, context: BatchContext, message: Union[TaskRequest, TaskResult, TaskException, TaskActionRequest, ChildPipeline]):
-        """
-        Runs a Flink Task
-
-        :param context: context object provided by Flink
-        :param message: the task input protobuf message
-        """
-        with TaskContext(context, self._egress_type_name, self._serialiser) as task_context:
-            
-            for handler in self._handlers:
-                if handler.can_handle(task_context, message):
-                    
-                    try:
-                        _log.info(f'Starting {task_context}')
-
-                        handler.handle_message(self, task_context, message)
-
-                        _log.info(f'Finished {task_context}')
-                        
-                    except Exception as ex:
-                        _log.error(f'Error invoking {task_context} - {ex}')
-                        self.fail(task_context, message, ex)
-                        
-                    finally:
-                        
-                        return
-            
-            _log.error(f'Unsupported message type {message.typed_value.typename}')
-
     @staticmethod
     def send(func, *args, **kwargs) -> PipelineBuilder:
         """
@@ -256,7 +240,7 @@ class FlinkTasks(object):
         pipeline_protos = context.pipeline_state.pipeline
 
         if pipeline_protos is not None:
-            return _Pipeline.from_proto(pipeline_protos, self._serialiser, self._events)
+            return _Pipeline.from_proto(pipeline_protos, self._serialiser, self._events, self._storage)
         else:
             raise ValueError(f'Missing pipeline for task_id - {context.get_task_id()}')
 
